//+------------------------------------------------------------------+
//|                                                      ZigZags.mqh |
//|                                         Copyright 2011, AlexSTAL |
//|                                           http://www.alexstal.ru |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, AlexSTAL"
#property link      "http://www.alexstal.ru"

#include <CustPrevCalculated.mqh>
#include <OutsideBar.mqh>

//+------------------------------------------------------------------+
//| Result of working of the ZZ calculation function                 |
//+------------------------------------------------------------------+
enum CPCZZResultCode
  {
   CPCZZRC_NotInitialized,             // Class is not initialized
   CPCZZRC_NoData,                     // Failed to receive data (for the external bar too)
   CPCZZRC_NotChanged,                 // Rays of ZZ has not been changed
   CPCZZRC_Changed                     // ZZ has been changed
  };

//+------------------------------------------------------------------+
//| Structure of one bar of the ZZ indicator                         |
//+------------------------------------------------------------------+
struct ZZBar
  {
   double UP, DN;                      // Buffers of the ZigZag indicator, must be stored separately
   OrderFormationBarHighLow OB;        // Buffer for caching of external bar
  };

//+------------------------------------------------------------------+
//| Class CCustZigZagPPC - example class for using of ZigZag PPC     |
//| http://www.mql5.com/en/code/263                                  |
//+------------------------------------------------------------------+
class CCustZigZagPPC
  {
protected:
   bool Initialized;                         // The variable for protection from a programming error of class initialization
   bool UsesExternalData;                    // Use external data
   // Main parameters
   //----------------------------
   int BarsLimit;                            // Limit of bars for calculation is set once during initialization of the class
   double iMinMotion;                        // Minimum movement of price in points on the zero bar for the indicator to be recalculated
   bool iUseSmallerTFforEB;                  // Use more precise algorithm of calculation of the order of from of the High/Low bar
   // Arrays for data storage
   //----------------------------
   double LastBarLastHigh;                   // Auxiliary variable for optimization of calculations
   double LastBarLastLow;                    // Auxiliary variable for optimization of calculations
   uint DataBarsCount;                       // The number of significant bars for that are stored in the array for calculation, cannot exceed BarsLimit
   uint DataBarsCalculated;                  // The number of bars that are not changed for sure
   uint DataStartInd;                        // Initial value of indexing in the array for calculation (for wraparound addressing)
   ZZBar Buf[];                              // Data array for the calculation itself
   uint ZZDataBarsCalculated;                // The number of calculated bars of the ZZZ indicator
   //----------------------------
   uchar iExtPeriod;                         // Number of bars for the calculation of extremums, cannot be less than 2
   double iMinAmplitude;                     // Minimum price distance between adjacent peak and valley (otherwise it is not registered)
   datetime TimeFirstExtBar;                 // Specific variable only for the PPC algorithm (to avoid start of calculation on the external bar)
   char _Comb(int i, double H, double L, double Fup, double Fdn);
   int _iHighest(int startPos);
   int _iLowest(int startPos);
   bool ZZDataBarsCuttingLeft;               // Set only to 'true' at each call of the PrepareData() procedure
   CCustPrevCalculated *CustPrevCalculated;
   // Auxiliary variables
   //----------------------------
   int __Digits;
   double __Point;
   uint PInd(uint I, bool AsSeries) {return( AsSeries?((DataStartInd+DataBarsCount-1-I) % BarsLimit):((DataStartInd+I) % BarsLimit) );}
public:
   // Constructor and destructor
   void CCustZigZagPPC() {Initialized = false;}
   void ~CCustZigZagPPC() {if (CustPrevCalculated != NULL) delete CustPrevCalculated;}
   bool ItIsInitialized() {return(Initialized);}
   // Manual initialization of the class
   bool Init(CCustPrevCalculated *CPC, string Instr, ENUM_TIMEFRAMES TF, int Limit, CPCHistorySynchMode HSM, uchar HS, uint HSMinute, uchar MinMotion, bool UseSmallerTFforEB, uchar ExtPeriod, uchar MinAmplitude);
   // Main procedure of synchronization of data
   CPCPrepareDataResultCode PrepareData(CPCPrepareDataResultCode resData);   // Procedure for loading new data for the calculation
   // Main procedure of calculation of the indicator
   CPCZZResultCode Calculate();
   // The function of access to the calculated data
   uint GetBarsCount() {return(DataBarsCount);}                // Returns the nu,ber of available bars
   uint GetBarsCalculated() {return(ZZDataBarsCalculated);}    // Returns the number of calculated bars
   double GetUP(uint shift, bool AsSeries);                    // Returns the value of the ZigZag peak for a bar
   double GetDN(uint shift, bool AsSeries);                    // Returns the value of the ZigZag valley for a bar
   OrderFormationBarHighLow GetOB(uint shift, bool AsSeries);  // Returns the Outside value of a bar
  };
//+------------------------------------------------------------------+
bool CCustZigZagPPC::Init(CCustPrevCalculated *CPC, string Instr, ENUM_TIMEFRAMES TF, int Limit, CPCHistorySynchMode HSM, uchar HS, uint HSMinute, uchar MinMotion, bool UseSmallerTFforEB, uchar ExtPeriod, uchar MinAmplitude)
{
   if (Initialized)
      return(false);
   
   if (CPC == NULL)
     {
      CustPrevCalculated = new CCustPrevCalculated;
      UsesExternalData = false;
     } else {
      CustPrevCalculated = CPC;
      UsesExternalData = true;
     }
   if (CustPrevCalculated.InitData(Instr, TF, Limit, HSM, HS, HSMinute))
     {
      if (ArrayResize(Buf, Limit) == -1)
         return(Initialized);

      BarsLimit = Limit;
      
      __Digits = (int)SymbolInfoInteger(Instr, SYMBOL_DIGITS);
      __Point = SymbolInfoDouble(Instr, SYMBOL_POINT);
      
      if (ExtPeriod >= 2)
         iExtPeriod = ExtPeriod;
      else
         iExtPeriod = 2;
      
      iMinAmplitude = NormalizeDouble(MinAmplitude * __Point, __Digits);
      
      if (MinMotion >= 1)
         iMinMotion = NormalizeDouble(MinMotion * __Point, __Digits);
      else
         iMinMotion = NormalizeDouble(__Point, __Digits);
      
      iUseSmallerTFforEB = UseSmallerTFforEB;
      
      Initialized = true;
      return(true);
     } else {
      Initialized = false;
      return(false);
     }
}
//+------------------------------------------------------------------+
CPCPrepareDataResultCode CCustZigZagPPC::PrepareData(CPCPrepareDataResultCode resData)
{
   if (!Initialized)
      return(CPCPDRC_NoData);
   
   // Prepare trade data for working
   CPCPrepareDataResultCode res = resData;
   DataBarsCount = CustPrevCalculated.GetDataBarsCount();
   DataBarsCalculated = CustPrevCalculated.GetDataBarsCalculated();
   DataStartInd = CustPrevCalculated.GetDataStartInd();
   
   // If the bars are cut from the left side, set the flag in the auxiliary variable
   // Only 'true' value is set, 'false' is zeroized on its own if necessary
   if (CustPrevCalculated.GetDataBarsCuttingLeft())
      ZZDataBarsCuttingLeft = true;
   
   // Zeroize the newly added or synchronized bars in the ZZ buffer
   for (uint i = (DataBarsCalculated == 0)?0:(DataBarsCalculated+1); i < DataBarsCount; i++)
     {
      Buf[PInd(i, false)].UP = EMPTY_VALUE;
      Buf[PInd(i, false)].DN = EMPTY_VALUE;
      Buf[PInd(i, false)].OB = OFBError;
     }
   
   // Move the calculated ZZ point if there's been a rollback of history
   ZZDataBarsCalculated = MathMin(DataBarsCalculated, ZZDataBarsCalculated);
   
   // Optimization of calculation at ticks on the last bar
   double LastHigh = CustPrevCalculated.GetDataHigh(DataBarsCount-1, false);
   double LastLow = CustPrevCalculated.GetDataLow(DataBarsCount-1, false);
   if (res == CPCPDRC_SynchOnlyLastBar)
     {
      if ( (NormalizeDouble(LastHigh-LastBarLastHigh, __Digits) >= iMinMotion) || (NormalizeDouble(LastBarLastLow-LastLow, __Digits) >= iMinMotion) )
        {
         LastBarLastHigh = LastHigh;
         LastBarLastLow = LastLow;
        } else {
         // Calculation won't be performed at this tick, since the price hasn't been changed withing
         // the zero bar or the difference doesn't exceed the minimal limit
         // determined by the MinMotion variable
         res = CPCPDRC_NoRecountNotRequired;
         DataBarsCalculated = DataBarsCalculated + 1;
        }
     } else {
      LastBarLastHigh = LastHigh;
      LastBarLastLow = LastLow;
     }
   
   return(res);
}
//+------------------------------------------------------------------+
double CCustZigZagPPC::GetUP(uint shift, bool AsSeries)
{
   if ( shift > (DataBarsCount-1) )
      return(EMPTY_VALUE);
   return(Buf[PInd(shift, AsSeries)].UP);
}
//+------------------------------------------------------------------+
double CCustZigZagPPC::GetDN(uint shift, bool AsSeries)
{
   if ( shift > (DataBarsCount-1) )
      return(EMPTY_VALUE);
   return(Buf[PInd(shift, AsSeries)].DN);
}
//+------------------------------------------------------------------+
OrderFormationBarHighLow CCustZigZagPPC::GetOB(uint shift, bool AsSeries)
{
   if ( shift > (DataBarsCount-1) )
      return(0);
   return(Buf[PInd(shift, AsSeries)].OB);
}
//+------------------------------------------------------------------+
//|  searching index of the highest bar                              |
//+------------------------------------------------------------------+
int CCustZigZagPPC::_iHighest(int startPos)
  {
   uchar depth = iExtPeriod;
   int index = startPos;
   //--- depth correction if need
   if ( (startPos-depth) < 0 )
      depth = (uchar)startPos;
   double max = CustPrevCalculated.GetDataHigh(startPos, false);
   //--- start searching
   double tmp;
   for (int i = startPos; i > (startPos-depth); i--)
     {
      tmp = CustPrevCalculated.GetDataHigh(i, false);
      if (tmp > max)
        {
         index = i;
         max = tmp;
        }
     }
   //--- return index of the highest bar
   return(index);
  }
//+------------------------------------------------------------------+
//|  searching index of the lowest bar                               |
//+------------------------------------------------------------------+
int CCustZigZagPPC::_iLowest(int startPos)
  {
   uchar depth = iExtPeriod;
   int index = startPos;
   //--- depth correction if need
   if ( (startPos-depth) < 0 )
      depth = (uchar)startPos;
   double min = CustPrevCalculated.GetDataLow(startPos, false);
   //--- start searching
   double tmp;
   for (int i = startPos; i > (startPos-depth); i--)
     {
      tmp = CustPrevCalculated.GetDataLow(i, false);
      if (tmp < min)
        {
         index = i;
         min = tmp;
        }
     }
   //--- return index of the lowest bar
   return(index);
  }
//+------------------------------------------------------------------+
//|                                     |
//+------------------------------------------------------------------+
char CCustZigZagPPC::_Comb(int i, double H, double L, double Fup, double Fdn)
{
   //----
   if (Fup==H && (Fdn==0 || (Fdn>0 && Fdn>L))) return(1);  //potential peak on the calculated bar
   if (Fdn==L && (Fup==0 || (Fup>0 && Fup<H))) return(-1); //potential valley on the calculated bar
   if (Fdn==L && Fup==H)                                   //potential peak and valley on the calculated bar 
     {
      OrderFormationBarHighLow OrderFormationHL = OFBError;
      // Attempt to find cached data in the buffer OB[]
      if (Buf[PInd(i, false)].OB == OFBError)
        {
         OrderFormationHL = GetOrderFormationBarHighLow(Symbol(), Period(), DataBarsCount - i - 1, iUseSmallerTFforEB);
         if (OrderFormationHL != OFBError)
            Buf[PInd(i, false)].OB = OrderFormationHL;
         else
            return(CHAR_MIN); // Failed to return data as a result of temporary unavailability of the data
        } else {
         OrderFormationHL = Buf[PInd(i, false)].OB;
        }
      
      switch(OrderFormationHL)
        {
         case OFBLowHigh:       //probably LOW was formed before HIGH (bullish bar)
            return(2);
            break;
         case OFBHighLow:       //probably HIGH was formed before LOW (bearish bar)
            return(-2); 
            break;
        }
     }
   //----  
   return(0);           //nothing is on the calculated bar...
}
//+------------------------------------------------------------------+
CPCZZResultCode CCustZigZagPPC::Calculate()
{
   if (!Initialized)
      return(CPCZZRC_NotInitialized);
   
   // If the data provider is not external, synchronize the history
   if (!UsesExternalData)
     {
      CPCPrepareDataResultCode resCPCPD = CustPrevCalculated.PrepareData();
      CPCPrepareDataResultCode resPD = PrepareData(resCPCPD);

      if (resPD == CPCPDRC_NoData)
         return(CPCZZRC_NoData);
      if (resPD == CPCPDRC_NoRecountNotRequired)
         return(CPCZZRC_NotChanged);
     }
   
   uint BarsForRecalculation;
   if ( DataBarsCalculated != 0 )
      BarsForRecalculation = DataBarsCount - ZZDataBarsCalculated - 1;
   else
      BarsForRecalculation = DataBarsCount - iExtPeriod - 1;
   
   // The history was cut from the left side
   if (ZZDataBarsCuttingLeft)
     {
      //    TimeFirstExtBar
      double tmpDN, tmpUP;
      uchar tmpC;
      for (uint i = 0; i < (DataBarsCount - BarsForRecalculation - 1); i++)
        {
         tmpC = 0;
         tmpDN = Buf[PInd(i, false)].DN;
         tmpUP = Buf[PInd(i, false)].UP;
         if (tmpDN != EMPTY_VALUE)
            tmpC = tmpC + 1;
         if (tmpUP != EMPTY_VALUE)
            tmpC = tmpC + 2;
         if ( tmpC == 3 )
           {
            Buf[PInd(i, false)].DN = EMPTY_VALUE;
            Buf[PInd(i, false)].UP = EMPTY_VALUE;
           }
         if ( (tmpC == 1) || (tmpC == 2) )
           { // Single first extremum is found
            TimeFirstExtBar = CustPrevCalculated.GetDataTime(i, false);
            break;
           }
        }
      
      ZZDataBarsCuttingLeft = false;
     }
   
   //======================================================
   //======== main loop ===============================
   //======================================================
   CPCZZResultCode res = CPCZZRC_NotChanged;
   
   char LET, tmpComb;
   double H, L, Fup, Fdn;
   uint lastUPbar, lastDNbar;
   double lastUP, lastDN;
   uint m, n; // For searching of the last extremum
   for (uint i = (DataBarsCount - BarsForRecalculation - 1); i < DataBarsCount; i++)
     {
      // Search of the last extremum
      // ---------------------------
      lastUP = 0;
      lastDN = 0;
      lastUPbar = i;
      lastDNbar = i;
      LET = 0; 
      m = 0; n = 0;
      while ( Buf[PInd(lastUPbar, false)].UP == EMPTY_VALUE )
        {
         if ( lastUPbar < iExtPeriod )
            break;
         lastUPbar--;
        }
      lastUP = Buf[PInd(lastUPbar, false)].UP; //probably, the last peak is found
      while ( Buf[PInd(lastDNbar, false)].DN == EMPTY_VALUE)
        {
         if ( lastDNbar < iExtPeriod )
            break;
         lastDNbar--;
        } 
      lastDN = Buf[PInd(lastDNbar, false)].DN; //probably, the last valley is found
   
      if ( lastUPbar > lastDNbar)
         LET = 1;
      if ( lastUPbar < lastDNbar)
         LET = -1;
      if ( lastUPbar == lastDNbar)
        {
         //lastUPbar==lastDNbar we need to find out which single extremum is the last one
         m = lastUPbar;
         n = m;
         while ( m == n )
           {
            m--; n--;
            while ( Buf[PInd(m, false)].UP == EMPTY_VALUE )
              {
               if ( m < iExtPeriod )
                  break;
               m--;
              } //probably, the last peak is found
            while ( Buf[PInd(n, false)].DN == EMPTY_VALUE )
              {
               if ( n < iExtPeriod )
                  break;
               n--;
              } //probably, the last valley is found
            if ( MathMax(m, n) < iExtPeriod )
               break;
           }
         if (m > n)
            LET = 1;       //base reference is a peak
         else
            if (m < n)
               LET = -1; //base reference is a valley
        }
      // if LET==0 then it's the reference point or an external bar with 2 extremums is registered in the very beginning
      // End of search of the last extremum
      // ----------------------------------

      //---- analyze price extremums for the calculated period:       
      H = CustPrevCalculated.GetDataHigh(_iHighest(i), false);
      L = CustPrevCalculated.GetDataLow(_iLowest(i), false);
      Fup = CustPrevCalculated.GetDataHigh(i, false);
      Fdn = CustPrevCalculated.GetDataLow(i, false);
      
      //---- analyze the situation and consider the possibility of registration of new extremums: 
      tmpComb = _Comb(i,H,L,Fup,Fdn);
      if (tmpComb == CHAR_MIN)
         return(CPCZZRC_NoData);
      switch(tmpComb)
        {
         //---- a potential peak on the calculated bar (Comb)      
         case 1 :
            switch(LET)
              {
               case 1 :
                  //previous extremums is also a peak, choose the biggest:
                  if ( lastUP < Fup )
                    {
                     Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                     Buf[PInd(i, false)].UP = Fup;
                     res = CPCZZRC_Changed;
                    }
                  break;
               case -1 :
                  if ( (Fup-lastDN) > iMinAmplitude ) //previous extremum is a valley
                    {
                     Buf[PInd(i, false)].UP = Fup;
                     res = CPCZZRC_Changed;
                    }
                  break; 
               default :
                  Buf[PInd(i, false)].UP = Fup;
                  TimeFirstExtBar = CustPrevCalculated.GetDataTime(i, false); //0 - it means that it's the start of calculation 
                  res = CPCZZRC_Changed;
                  break; 
              }
            break;

         //----  a potential valley on the calculated bar (Comb)          
         case -1 :
            switch(LET)
              {
               case 1 :
                  if ( (lastUP-Fdn) > iMinAmplitude ) //previous extremum is a peak
                    {
                     Buf[PInd(i, false)].DN = Fdn;
                     res = CPCZZRC_Changed;
                    }
                  break; 
               case -1 :
                  //previous extremum is also a valley, choose the lowest:
                  if ( lastDN > Fdn )
                    {
                     Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                     Buf[PInd(i, false)].DN = Fdn;
                     res = CPCZZRC_Changed;
                    }
                  break;
               default :
                  Buf[PInd(i, false)].DN = Fdn;
                  TimeFirstExtBar = CustPrevCalculated.GetDataTime(i, false); //0 - it means that it's the start of calculation 
                  res = CPCZZRC_Changed;
                  break;
              }
            break;

         //---- potential peak and valley on the calculated bar (Comb)        
         case 2 : //probably LOW was formed before HIGH (bullish bar)
            switch(LET)
              {
               case 1 : //previous extremum is a peak
                  if ( (Fup-Fdn) > iMinAmplitude )
                    {
                     if ( (lastUP-Fdn) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].UP = Fup;
                        Buf[PInd(i, false)].DN = Fdn;
                        res = CPCZZRC_Changed;
                       } else {
                        if ( lastUP < Fup )
                          {
                           Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                           Buf[PInd(i, false)].UP = Fup;
                           res = CPCZZRC_Changed;
                          }
                       }
                    } else {
                     if ( (lastUP-Fdn) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].DN = Fdn;
                        res = CPCZZRC_Changed;
                       } else {
                        if ( lastUP < Fup )
                          {
                           Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                           Buf[PInd(i, false)].UP = Fup;
                           res = CPCZZRC_Changed;
                          }
                       }
                    }
                  break;
               case -1 : //previous extremum is a valley
                  if ( (Fup-Fdn) > iMinAmplitude )
                    {
                     Buf[PInd(i, false)].UP = Fup;
                     if ( (Fdn < lastDN) && (CustPrevCalculated.GetDataTime(lastDNbar, false) > TimeFirstExtBar) )
                       {
                        Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                        Buf[PInd(i, false)].DN = Fdn;
                       }
                     res = CPCZZRC_Changed;
                    } else {
                     if ( (Fup-lastDN) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].UP = Fup;
                        res = CPCZZRC_Changed;
                       } else {
                        if ( lastDN > Fdn )
                          {
                           Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                           Buf[PInd(i, false)].DN = Fdn;
                           res = CPCZZRC_Changed;
                          }
                       }
                    }
                  break;
               default: break;
              } //switch LET
            break;

         //---- potential peak and valley on the calculated bar (Comb)        
         case -2 : //probably HIGH was formed before LOW (bearish bar)
            switch(LET)
              {
               case 1 : //previous extremum is a peak
                  if ( (Fup-Fdn) > iMinAmplitude )
                    {
                     Buf[PInd(i, false)].DN = Fdn;
                     if ( (lastUP < Fup) && (CustPrevCalculated.GetDataTime(lastUPbar, false) > TimeFirstExtBar) )
                       {
                        Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                        Buf[PInd(i, false)].UP = Fup;
                       }
                     res = CPCZZRC_Changed;
                    } else {
                     if ( (lastUP-Fdn) > iMinAmplitude)
                       {
                        Buf[PInd(i, false)].DN = Fdn;
                        res = CPCZZRC_Changed;
                       } else {
                        if ( lastUP < Fup )
                          {
                           Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                           Buf[PInd(i, false)].UP = Fup;
                           res = CPCZZRC_Changed;
                          }
                       }
                    }
                  break;
               case -1 : //previous extremum is a valley
                  if ( (Fup-Fdn) > iMinAmplitude )
                    {
                     if ( (Fup-lastDN) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].UP = Fup;
                        Buf[PInd(i, false)].DN = Fdn;
                        res = CPCZZRC_Changed;
                       } else {
                        if (lastDN > Fdn)
                          {
                           Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                           Buf[PInd(i, false)].DN = Fdn;
                           res = CPCZZRC_Changed;
                          }
                       }
                    } else {
                     if ( (Fup-lastDN) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].UP = Fup;
                     res = CPCZZRC_Changed;
                       } else {
                        if ( lastDN > Fdn)
                          {
                           Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                           Buf[PInd(i, false)].DN = Fdn;
                           res = CPCZZRC_Changed;
                          }
                       }
                    }
                  break;
               default: break;
              } //switch LET
            break;
         
         default: break;
        } // switch ()
      ZZDataBarsCalculated = i + 1; // right here, because we can exit earlier
     } // for (main)
   
   return(res);
}